/* IluJava_IluCall.c */
/* Chris Jacobi, April 28, 1997 4:12 pm PDT */

/*
 * Copyright (c) 1996, 1997 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluCall.c,v 1.58 1997/04/28 23:39:37 jacobi Exp $ */

/*
 * C side for IluCall.java
 */


#include "IluJava_CallMacros.h"
#include "IluJava_JMon.h"


/* The CALLOCK is s static global lock.
 * Never enter any other locks while CALLOCK is held.
 */
static ILUJAVA_JMON_PTR callLock = 0;
#define ENTER_CALLOCK() ILUJAVA_MON_ENTER(callLock)
#define EXIT_CALLOCK() ILUJAVA_MON_EXIT(callLock)

static void InitCallLock()
{
    callLock = ILUJAVA_MON_ALLOC();       
}



/*
 * If only correct stubs were used, unhand(jjcall)->needFinish
 * wouldn't need to be locked. Correct stubs would guarantee
 * sequential use and finish call before finalization.
 *
 * However, there are bad stubs, and programming errors which
 * are just not worth spending all the time thinking about. 
 */
 
static void SetNeedFinish(JIluCall jjcall)
{
   ENTER_CALLOCK();
   unhand(jjcall)->needFinish = 1;
   EXIT_CALLOCK();
}

static Jint GetClearNeedFinish(JIluCall jjcall)
/* Actual ilu_FinishCall is conditional and monitored because we 
 * don't really trust the stubs (in case of applets); (nor do we
 * really trust our own error handling to be really correct).
 */
{
   Jint nf;
   ENTER_CALLOCK();
   nf = unhand(jjcall)->needFinish;
   unhand(jjcall)->needFinish = 0;
   EXIT_CALLOCK();
   return nf;
}



static int ilujava_nullBuffer = 0; 
    /* placeholder where empty pointers can point to */

static JString ilujava_jEmptyString;
    /* Replace null strings by this empty string 
     * Initialized in "givePrototype"
     */

static /*interpreter.h*/ ClassClass * ilujava_javaClassIluCall;

/* called from static block to initialize the J-class */
EXPORTJAVA void 
xerox_ilu_IluCall_givePrototype(JIluCall jjcall)
{
    if (callLock==0) InitCallLock();
    ilujava_javaClassIluCall = obj_classblock(jjcall);
    ilujava_jEmptyString = makeJavaString((char *)&ilujava_nullBuffer, 0);
}


/* Create an object */
/* Made native because I want to add more fields as well as add caching */
/* Consider making this available to java ? */
static JIluCall 
ilujava_constructIluCall(JIluCall unused)
{
    JIluCall jjcall;
    jjcall = (JIluCall) ObjAlloc(ilujava_javaClassIluCall, 0);
    	/* ObjAlloc does not execute Java constructor */
    xerox_ilu_IluCall_initCallShell(jjcall);
    return jjcall;
}


EXPORTJAVA void 
xerox_ilu_IluCall_nativeFinalize(JIluCall jjcall)
/* Called by the java finalize procedure.
 * Not monitored as long as there is no other 
 * call.  
 */
{
    ilu_Call cCall;
    TRANSITION(CALL_FINALIZED);
    if (unhand(jjcall)->needFinish) {
        ilu_DebugPrintf("$ERROR: IluCall_nativeFinalize not finished\n");
    }
    cCall = Access(jjcall, yCall, ilu_Call);
    if (cCall) {
        Access(jjcall, yCall, ilu_Call) = 0;
        ilu_free(cCall);
    }
}


EXPORTJAVA void
xerox_ilu_IluCall_initCallShell(JIluCall jjcall)
    /* Native side for makeCallShell.
     * Trusts "jjcall" to be freshly allocated and therefore may 
     * avoid monitor locking.
     */
{
    ilu_Call cCall = ilu_must_malloc(sizeof(struct _ilu_Call_s));
    TRANSITION(CALL_INACTIVE);
    Access(jjcall, yCall, ilu_Call) = cCall;
}


EXPORTJAVA JIluCall 
xerox_ilu_IluCall_nativeStartCall(
	JIluCall junused,
	struct Hxerox_ilu_IluMethodRep * jjMethodRep,
	struct Hxerox_ilu_IluSurrogateObject * jjDiscriminantObj,
	struct Hxerox_ilu_IluPassport * jjPassport
	)
/* (surrogate) stub initiating a call */
{
    ILU_ERRS((IoErrs)) errs;
    ilu_Connection newConnection;
    ilu_boolean success;
    JIluCall jjcall;
    ilu_Call cCall;
    ilu_Server cServer;
    JIluOInt jjoi;
    JIluClassRep jjStaticClass = 
    	unhand(jjMethodRep)->jjClassRep; /*methods carry static class*/
     
    jjcall = ilujava_constructIluCall(junused);
    cCall = Access(jjcall, yCall, ilu_Call);
    /* GC: stuffing features of the following java objects into the C 
     * heap is ok:
     *   jObj is is held on by stub;
     *   jjMethodRep, jjClassRep are not subject to GC
     */
    
    ILUJAVA_EXPECT(CALL_INACTIVE, 0)
    unhand(jjcall)->jDontGCcall1 = (HObject*) jjDiscriminantObj;
        /* Semi paranoya: we are also saving the internal object which has
         * a reference to the object itself; by as well saving it directly
         * we prevent gc even in case of nasty accident to internal object.
         * Never mind whether such accidents are possible or not
         */
    unhand(jjcall)->jjMethodRep = jjMethodRep;
    unhand(jjcall)->jjClassRep = jjStaticClass;
    unhand(jjcall)->jMustFork = 0;
    jjoi = xerox_ilu_IluOInt_nPeekWithCall(
        0, (JIluObject) jjDiscriminantObj, jjcall);
    unhand(jjcall)->jDontGCcall2 = (HObject*) jjoi;
    if (jjoi == 0) {
        unhand(jjcall)->jMustFork = 0;
        _ilujava_IluErrorToCallException(0, "object probably destroyed");
        /* This happens only when object has been garbage collected: impossible.
         * Or if for some future version of ilu explicit destruction of         
         * objects gets supported. 
         */ 
        return jjcall;
    }
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_StartCall j:%x c:%x\n", jjcall, cCall);
    }
    cServer = _ilujava_ServerEnterHolds(jjoi);
    if (cServer == 0) {
        _ilujava_IluErrorToException(0, "ilu_StartCall without server");
        return jjcall;
    }
    success = ilu_StartCall(
    	cCall,
    	cServer, 
        Access(jjStaticClass, yIluClass, ilu_Class),
	Access(jjMethodRep, yIluMethod, ilu_Method),
	JAVALANGIDX,
	_ilujava_getIluPassport((ilu_refany) jjPassport), 
	&newConnection, 
	&errs
    	);
    _ilujava_ServerExitHolds(cServer);
    if (newConnection) {
        if (_ilujava_callsFlag>1) {
            ilu_DebugPrintf("$ IluCall_StartCall: new connection %x\n", 
                newConnection);
        }
        Access(jjcall, yCallSurrConn, ilu_Connection) = newConnection;
        unhand(jjcall)->jMustFork = 1;
    }
    if (success) {
        TRANSITION(CALL_SZ)
        SetNeedFinish(jjcall);
        unhand(jjcall)->jNeedsSizing = (Jboolean) ilu_CallNeedsSizing(cCall);
        if (_ilujava_callsFlag>1) {
            ilu_DebugPrintf("$ IluCall_StartCall success j:%x\n", jjcall);
        }
        return jjcall;
    } else {
        if (_ilujava_callsFlag>1) {
            ilu_DebugPrintf("$ IluCall_StartCall failure j:%x\n", jjcall);
        }
        _ilujava_IluErrorToException(&errs, "failed ilu_StartCall");
        return jjcall;
    }
} /* nativeStartCall */


static void 
finishCall_DontRaiseErrors0(JIluCall jjcall)
/* Like xerox_ilu_IluCall_finishCall but for internal use in 
 * cases where errors are reported anyway and we don't
 * want to report more or different errors.
 *
 * Do the transition self checking
 */
{
    Jint needF = GetClearNeedFinish(jjcall);
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_finishCall_DRE0 j:%x, n:%d\n", jjcall, needF);
    }
    if (needF) {
        ILU_ERRS((bad_locks, IoErrs)) err = ILU_INIT_NO_ERR;
        ilu_Call cCall;
        EXPECT_UNFINISHED_QUIET()
        cCall = Access(jjcall, yCall, ilu_Call);
        if (cCall) ilu_FinishCall(cCall, &err);
        /* ignore ilu errors  */
        unhand(jjcall)->jDontGCshort = 0;
        unhand(jjcall)->jDontGCcall1 = 0;
        unhand(jjcall)->jDontGCcall2 = 0;
        TRANSITION(CALL_FINISHED)
    }
}


static void 
finishCall_DontRaiseErrors1(JIluCall jjcall)
/* Like finishCall_DontRaiseErrors0 but doesn't deal with
 * self checking.
 */
{
    Jint needF = GetClearNeedFinish(jjcall);
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_finishCall_DRE1 j:%x, n:%d\n", jjcall, needF);
    }
    if (needF) {
        ilu_Call cCall;
        ILU_ERRS((bad_locks, IoErrs)) err = ILU_INIT_NO_ERR;
        cCall = Access(jjcall, yCall, ilu_Call);
        if (cCall) ilu_FinishCall(cCall, &err);
        /* ignore ilu errors  */
        unhand(jjcall)->jDontGCshort = 0;
        unhand(jjcall)->jDontGCcall1 = 0;
        unhand(jjcall)->jDontGCcall2 = 0;
    }
}


EXPORTJAVA void 
xerox_ilu_IluCall_finishCall(JIluCall jjcall)
{
    Jint needF = GetClearNeedFinish(jjcall);
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_finishCall j:%x, n:%d\n", jjcall, needF);
    }
    /* don't check state as this is used in try-finally */
    if (needF) {
        ilu_Call cCall;
        ILU_ERRS((bad_locks, IoErrs)) err = ILU_INIT_NO_ERR;
        cCall = Access(jjcall, yCall, ilu_Call);
        if (cCall) {
            ilu_FinishCall(cCall, &err);
            if (ILU_ERRNOK(err)) {
                /* GRRRR; it would be nice to know if we are in the 
                 * try-finally because of an error or not
                 */
                _ilujava_IluErrorToException(&err, "finishCall"); 
            }
        }
        unhand(jjcall)->jDontGCshort = 0;
        unhand(jjcall)->jDontGCcall1 = 0;
        unhand(jjcall)->jDontGCcall2 = 0;
    }
    TRANSITION(CALL_FINISHED)
}


EXPORTJAVA void
xerox_ilu_IluCall_startWriteRequest(JIluCall jjcall, Jint sz)
    /* used by surrogates */
{
    ilu_Call cCall;
    ILU_ERRS((IoErrs)) err;
    EXPECT_SZ_OR_INBETWEEN(ILUJAVA_VOID)
    cCall = Access(jjcall, yCall, ilu_Call);
    if (!ilu_StartRequest(cCall, (ilu_cardinal) sz, &err)) {
        /* error */
        finishCall_DontRaiseErrors0(jjcall);
        _ilujava_IluErrorToException(&err, "startWriteRequest");
        return;
    }
    TRANSITION(CALL_OUT)
}


EXPORTJAVA void
xerox_ilu_IluCall_doneWriteRequest(JIluCall jjcall)
    /* used by surrogates */
{
    ILU_ERRS((IoErrs)) err;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_OUT, ILUJAVA_VOID)
    if (!ilu_FinishRequest(cCall, &err)) {
        /* Error. 
         * Call is over; call into java anyway in case data structures
         * would need to be returned.
         */
        finishCall_DontRaiseErrors0(jjcall);
        _ilujava_IluErrorToException(&err, "doneWriteRequest");
        return;
    }
    TRANSITION(CALL_INBETWEEN)
    if (cCall->ca_method->me_asynchronous) {
        finishCall_DontRaiseErrors1(jjcall);
    }
    return;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_startReadReply(JIluCall jjcall)
    /* used by surrogates */
{
    ilu_cardinal errorStat;
    ilu_ProtocolException protoErr;
    ILU_ERRS((bad_locks, IoErrs)) err = ILU_INIT_NO_ERR;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_INBETWEEN, 0)
    protoErr = ilu_GetReply(cCall, &errorStat, &err);
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_startReadReply proto: %d user: %d\n", 
            (int) protoErr, (int) errorStat);
    }
    if (protoErr == ilu_ProtocolException_Success) {
        TRANSITION(CALL_IN)
        return errorStat;
    } else {
        int asint = (int) protoErr;
        finishCall_DontRaiseErrors0(jjcall);
        asint = - asint; /*this should make it negative*/
        TRANSITION(CALL_FINISHED)
        if (protoErr == ilu_ProtocolException_Not) {
            _ilujava_IluErrorToException(&err, "ProtocolException");
        }
        return (Jint) asint;
    }
} /*startReadReply*/


EXPORTJAVA void
xerox_ilu_IluCall_doneReadReply(JIluCall jjcall)
/* used by surrogates */
{
    ILU_ERRS((IoErrs)) err;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    if (unhand(jjcall)->needFinish) {
        ILUJAVA_EXPECT(CALL_IN, ILUJAVA_VOID)
        if ( ! ilu_ReplyRead(cCall, &err)) {
            /* Error; Call finishCall anyway. */
            _ilujava_Report(&err, "doneReadReply"); 
            finishCall_DontRaiseErrors0(jjcall);
            _ilujava_IluErrorToException(&err, "doneReadReply");
            return;
        }
    }
    TRANSITION(CALL_INBETWEEN)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_startReadRequest(JIluCall jjcall)
    /* used by true objects */
{
    TRANSITION(CALL_IN)
    /* I think there is nothing to do */
    return 0;
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_doneReadRequest(JIluCall jjcall)
    /* used by true objects */
{
    ILU_ERRS((IoErrs)) err;
    ilu_boolean     ans;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_IN, 0)
    ans = ilu_RequestRead(cCall, &err);
    if (ILU_ERRNOK(err)) {
        finishCall_DontRaiseErrors0(jjcall);
    }
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


#define hasExceptions(jjcall) ((unhand(unhand(jjcall)->jjMethodRep)->jjExceptions)!=0)

EXPORTJAVA Jint
xerox_ilu_IluCall_beginSizingReply(JIluCall jjcall)
    /* used by true objects */
{
    ILU_ERRS((IoErrs)) err;
    ilu_cardinal ret; 
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_INBETWEEN, 0)
    ret = ilu_BeginSizingReply(cCall, hasExceptions(jjcall), &err);
    TRANSITION(CALL_SZ)
    return (Jint) ret;
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_startWriteReply(JIluCall jjcall, Jint argSize)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err = ILU_INIT_NO_ERR;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    EXPECT_SZ_OR_INBETWEEN(0)
    ans = ilu_BeginReply(
        cCall, hasExceptions(jjcall), (ilu_cardinal) argSize, &err);
    if (! ans) {
        finishCall_DontRaiseErrors1(jjcall);
    }
    TRANSITION(CALL_OUT)
    return (ans!=0);
}

EXPORTJAVA Jint
xerox_ilu_IluCall_beginSizingException(JIluCall jjcall, Jint eindex)
    /* used by true objects */
{
    ILU_ERRS((IoErrs)) err;
    ilu_cardinal ret; 
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_INBETWEEN, 0)
    ret = ilu_BeginSizingException(cCall, (ilu_integer) eindex, &err);
    TRANSITION(CALL_SZ)
    return (Jint) ret;
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_startWriteException(JIluCall jjcall, Jint evalu, Jint argSize)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall;
    cCall = Access(jjcall, yCall, ilu_Call);
    
    EXPECT_SZ_OR_INBETWEEN(0)
    TRANSITION(CALL_OUT)
    ans = ilu_BeginException(
        cCall, (ilu_integer) evalu, (ilu_cardinal) argSize, &err);
    if (! ans) {
        finishCall_DontRaiseErrors1(jjcall);
    }
    return (ans!=0);
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_noReply(JIluCall jjcall)
    /* used by true objects */
{
    ILU_ERRS((bad_param, bad_locks, broken_locks)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_INBETWEEN, 0)
    ans = ilu_NoReply(cCall, &err);
    finishCall_DontRaiseErrors1(jjcall);
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_doneWriteReply(JIluCall jjcall)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_OUT, 0)
    ans = ilu_FinishReply(cCall, &err);
    finishCall_DontRaiseErrors1(jjcall);
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_doneWriteException(JIluCall jjcall)
    /* used by true objects */
{
    ILU_ERRS((bad_locks, IoErrs)) err;
    ilu_boolean ans; 
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ILUJAVA_EXPECT(CALL_OUT, 0)
    ans = ilu_FinishException(cCall, &err);
    TRANSITION(CALL_INBETWEEN)
    return (ans!=0);
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szInt32(JIluCall jjcall, Jint ival)
{
    PROLOG_SZ()
    sz = ilu_SizeOfInteger(cCall, (ilu_integer) ival, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outInt32(JIluCall jjcall, Jint ival)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputInteger(cCall, (ilu_integer) ival, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_inInt32(JIluCall jjcall)
{
    ilu_integer value;
    PROLOG(CALL_IN, 0)
    ilu_InputInteger(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jint) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szBool(JIluCall jjcall, Jboolean ival)
{
    PROLOG_SZ()
    sz = ilu_SizeOfBoolean(cCall, (ival != 0), &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outBool(JIluCall jjcall, Jboolean ival)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputBoolean(cCall, (ival != 0), &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_inBool(JIluCall jjcall)
{
    ilu_boolean value;
    PROLOG(CALL_IN, 0)
    ilu_InputBoolean(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jboolean) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szInt16(JIluCall jjcall, Jshort ival)
{
    PROLOG_SZ()
    sz = ilu_SizeOfShortInteger(cCall, (ilu_shortinteger) ival, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outInt16(JIluCall jjcall, Jshort ival)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputShortInteger(cCall, (ilu_shortinteger) ival, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jshort
xerox_ilu_IluCall_inInt16(JIluCall jjcall)
{
    ilu_shortinteger value;
    PROLOG(CALL_IN, 0)
    ilu_InputShortInteger(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jshort) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szByte(JIluCall jjcall, Jbyte ival)
{
    PROLOG_SZ()
    sz = ilu_SizeOfByte(cCall, (ilu_byte) ival, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outByte(JIluCall jjcall, Jbyte ival)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputByte(cCall, (ilu_byte) ival, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jbyte
xerox_ilu_IluCall_inByte(JIluCall jjcall)
{
    ilu_byte value;
    PROLOG(CALL_IN, 0)
    ilu_InputByte(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jbyte) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szChar16(JIluCall jjcall, Jchar ival)
{
    PROLOG_SZ()
    sz = ilu_SizeOfCharacter(cCall, (ilu_character) ival, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outChar16x(JIluCall jjcall, Jchar ival)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputCharacter(cCall, (ilu_character) ival, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jchar
xerox_ilu_IluCall_inChar16(JIluCall jjcall)
{
    ilu_character value;
    PROLOG(CALL_IN, 0)
    ilu_InputCharacter(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jchar) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szChar8(JIluCall jjcall, Jchar ival)
{
    PROLOG_SZ()
    sz = ilu_SizeOfShortCharacter(cCall, (ilu_shortcharacter) ival, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outChar8x(JIluCall jjcall, Jchar ival)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputShortCharacter(cCall, (ilu_shortcharacter) ival, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jchar
xerox_ilu_IluCall_inChar8(JIluCall jjcall)
{
    ilu_shortcharacter value;
    PROLOG(CALL_IN, 0)
    ilu_InputShortCharacter(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jchar) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szObjectx(JIluCall jjcall,
	JIluOInt jjoi, 
	Jboolean discriminant,
	JIluClassRep jjStaticClass
	)
{  
    JIluClassRep jjActualClass;
    ilu_Object cObject = 0;
    ilu_Class cStaticClass = 0;
    ilu_Class cActualClass = 0;
    ilu_Server cServer = 0;
    PROLOG_SZ()
    if (jjStaticClass==0) {
        _ilujava_IluErrorToCallException(0, "xxx-21a");
        return 0;
    }
    cStaticClass = Access(jjStaticClass, yIluClass, ilu_Class);
    if (jjoi) {
        jjActualClass = unhand(jjoi)->jjClassRep;
        if (jjActualClass==0) {
            _ilujava_IluErrorToCallException(0, "xxx-23a");
            return 0;
        }
        cActualClass = Access(jjActualClass, yIluClass, ilu_Class);
        /* Use Object's server; not call's server */
        cServer = _ilujava_EnterServer(jjoi, cActualClass);
        if (cServer==0) {
            _ilujava_IluErrorToCallException(0, "xxx-23b");
            return 0;
        }
        cObject = Access(jjoi, yIluKernelObject, ilu_Object); 
        if (cObject==0) {
            /*we freed the object right before entering the monitor*/
            ilu_ExitServer(cServer, cActualClass);
            _ilujava_IluErrorToCallException(0, "xxx-23b");
            return 0;
        }
    }
    sz = ilu_SizeOfObjectID(
        cCall, cObject, discriminant!=0, cStaticClass, &ioerrs);
    if (jjoi) ilu_ExitServer(cServer, cActualClass); 
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outObjectx(JIluCall jjcall,
	JIluOInt jjoi, 
	Jboolean discriminant,
	JIluClassRep jjStaticClass)
{
    JIluClassRep jjActualClass;
    ilu_Object cObject = 0;
    ilu_Class cStaticClass = 0;
    ilu_Class cActualClass = 0;
    ilu_Server cServer;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if (jjStaticClass==0) {
        _ilujava_IluErrorToCallException(&ioerrs, "outObject: no static class");
    }
    if (jjoi) {
        jjActualClass = unhand(jjoi)->jjClassRep;
        if (jjActualClass==0) {
            _ilujava_IluErrorToCallException(0, "xxx-23d");
            return;
        }
        /* Use Object's server; not call's server */
        cActualClass = Access(jjActualClass, yIluClass, ilu_Class);
        cServer = _ilujava_EnterServer(jjoi, cActualClass);
        if (cServer==0) {
            _ilujava_IluErrorToCallException(0, "xxx-23f bad server");
            return;
        }
        cObject = Access(jjoi, yIluKernelObject, ilu_Object);
        if (cObject==0) {
            /* we freed the object right before entering the monitor*/
            ilu_ExitServer(cServer, cActualClass);
            _ilujava_IluErrorToCallException(0, 
                "xxx-23e Object has been unregistered");
            return;
        }
    }
    if (_ilujava_callsFlag>0) {
       ilu_DebugPrintf("$ IluCall_outObjectx c<%x>\n", cObject);
    }
    cStaticClass = Access(jjStaticClass, yIluClass, ilu_Class);
    ilu_OutputObjectID(cCall, cObject, discriminant!=0, cStaticClass, &ioerrs);
    /* ?? I believe the locking comment means that doing nothing is the 
     * right thing to do ??
     */
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA JIluOInt
xerox_ilu_IluCall_inOI(JIluCall jjcall,
	Jboolean discriminant, 
	JIluClassRep jjStaticClass)
{
    JIluOInt jjoi; 
    ilu_Server cServer;
    ilu_Object cIluObject;
    ilu_Class cStaticClass;
    PROLOG(CALL_IN, 0)
    if (_ilujava_callsFlag>1) {
        ilu_DebugPrintf("$ IluCall_inOI\n");
    }
    if (jjStaticClass==0) {
        _ilujava_IluErrorToCallException(&ioerrs, "inOI: no static class");
    }
    cStaticClass = Access(jjStaticClass, yIluClass, ilu_Class);
    ilu_InputObjectID(
        cCall, &cIluObject, discriminant!=0, cStaticClass, &ioerrs);
    /*o!=NIL => Inside(*o's server...); ILU_ERRNOK(*err) => *o==NIL*/ 
    ERRCHECK(0)
    if (_ilujava_callsFlag>0) {
       ilu_DebugPrintf("$ IluCall_inOI got c<%x>\n", cIluObject);
    }
    if (cIluObject==0) {
        return 0;
    }
    cServer = ilu_ServerOfObject(cIluObject); /* inside server lock ! */
    jjoi = (JIluOInt) ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
    if (_ilujava_callsFlag>0) {
        ilu_DebugPrintf("$ IluCall_inOI: got j<%x>\n", jjoi);
    }
    if (jjoi == 0) {
        /* Object is not/no-more known; create a surrogate */
        if (_ilujava_callsFlag>0) {
            ilu_DebugPrintf("$ IluCall_inOI: needs to create ob\n");
        }
        if (ilu_TrueServerP(cServer)) {
            ilu_ExitServer(cServer, cStaticClass);
            _ilujava_IluErrorToCallException(0, "True object has disappeared");
            return 0;
        } else {
            jjoi = (JIluOInt) _ilujava_creatSurrOIFromRegistry(
            	cIluObject, (ilu_refany) jjStaticClass);
        }
    } else {
        if (unhand(jjoi)->retained != 0) {unhand(jjoi)->ghost = 1;}
    }
    unhand(jjcall)->jDontGCshort = (HObject*) jjoi; /*make traceable*/
    /* Need object's server; not call's server */
    ilu_ExitServer(cServer, cStaticClass); /*iluxport.h says: "static_type"*/
    return jjoi;
}

EXPORTJAVA Hxerox_ilu_IluOInt *
xerox_ilu_IluCall_getCallSingletonOIx(JIluCall jjcall)
{
    ilu_Server cServer;
    ilu_Object cIluObject;
    JIluOInt jjoi = 0; 
    PROLOG(CALL_IN, 0);
    cIluObject = ilu_GetCallSingleton(cCall, &ioerrs);
    ERRCHECK(0)
    if (cIluObject!=0) {
        cServer = ilu_ServerOfObject(cIluObject);
        jjoi = (struct Hxerox_ilu_IluOInt*)
    		ilu_GetLanguageSpecificObject(cIluObject, JAVALANGIDX);
        if (jjoi == 0) {
            ilu_ExitServer(cServer, ilu_IntroTypeOfCall(cCall));
            _ilujava_IluErrorToCallException(0, "yyy-2");
            return 0;
        }
        unhand(jjcall)->jDontGCshort = (HObject*) jjoi; /*make traceable*/
        ilu_ExitServer(cServer, ilu_IntroTypeOfCall(cCall));
        /*see ilu.c to figure out whether to use static or actual class...*/
        }
    return jjoi;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szOptional(JIluCall jjcall, Jboolean stat)
{
    ilu_boolean istat = (stat != 0);
    PROLOG_SZ()
    sz = ilu_SizeOfOptional(cCall, istat, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outOptional(JIluCall jjcall, Jboolean stat)
{
    ilu_boolean istat = (stat != 0);
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputOptional(cCall, istat, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jboolean
xerox_ilu_IluCall_inOptional(JIluCall jjcall)
{
    ilu_boolean optionalStatus;
    PROLOG(CALL_IN, 0)
    ilu_InputOptional(cCall, &optionalStatus, &ioerrs);
    ERRCHECK(0)
    return (Jboolean) optionalStatus;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szArray(JIluCall jjcall, Jint elementCount)
{
    ilu_cardinal length = (ilu_cardinal) elementCount;
    PROLOG_SZ()
    sz = ilu_SizeOfArray(cCall, length, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outArray(JIluCall jjcall, Jint elementCount)
{
    ilu_cardinal length = (ilu_cardinal) elementCount;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputArray(cCall, length, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_inArray(JIluCall jjcall)
{
    PROLOG(CALL_IN, ILUJAVA_VOID)
    ilu_InputArray(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_endArray(JIluCall jjcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    EXPECT_LIFE()
    x = ilu_EndArray(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szRecord(JIluCall jjcall)
{
    PROLOG_SZ()
    sz = ilu_SizeOfRecord(cCall, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outRecord(JIluCall jjcall)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputRecord(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_inRecord(JIluCall jjcall)
{
    PROLOG(CALL_IN, ILUJAVA_VOID)
    ilu_InputRecord(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_endRecord(JIluCall jjcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    EXPECT_LIFE()
    x = ilu_EndRecord(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szSequence(JIluCall jjcall, Jint leng, Jint limit)
{
    ilu_cardinal len = (ilu_cardinal) leng;
    ilu_cardinal lim = (ilu_cardinal) limit;
    PROLOG_SZ()
    sz = ilu_SizeOfSequence(cCall, len, lim, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outSequence(JIluCall jjcall, Jint leng, Jint limit)
{
    ilu_cardinal len = (ilu_cardinal) leng;
    ilu_cardinal lim = (ilu_cardinal) limit;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputSequence(cCall, len, lim, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_outSequenceMark(JIluCall jjcall, Jint extent)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputSequenceMark(cCall, (ilu_cardinal) extent, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_inSequence(JIluCall jjcall, Jint limit)
{
    ilu_cardinal len;
    ilu_cardinal lim = (ilu_cardinal) limit;
    PROLOG(CALL_IN, 0)
    ilu_InputSequence(cCall, &len, lim, &ioerrs);
    ERRCHECK(0)
    return (Jint) len;
}


EXPORTJAVA void
xerox_ilu_IluCall_inSequenceMark(JIluCall jjcall, Jint extent)
{
    PROLOG(CALL_IN, ILUJAVA_VOID)
    ilu_InputSequenceMark(cCall, (ilu_cardinal) extent, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_endSequence(JIluCall jjcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    EXPECT_LIFE()
    x = ilu_EndSequence(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szUnion(JIluCall jjcall, Jint discrim, Jint discrimSize)
{
    PROLOG_SZ()
    sz = ilu_SizeOfUnion(
        cCall, (ilu_cardinal) discrim, (ilu_cardinal) discrimSize, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outUnion(JIluCall jjcall, Jint discrim, Jint discrimSize)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputUnion(
        cCall, (ilu_cardinal) discrim, (ilu_cardinal) discrimSize, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_inUnion(JIluCall jjcall, Jint discrimSize)
{
    ilu_cardinal discriminator;
    PROLOG(CALL_IN, 0)
    ilu_InputUnion(cCall, &discriminator, (ilu_cardinal) discrimSize, &ioerrs);
    ERRCHECK(0)
    return (Jint) discriminator;
}


EXPORTJAVA void
xerox_ilu_IluCall_endUnion(JIluCall jjcall)
{
    ILU_ERRS((IoErrs)) ioerrs;
    ilu_boolean x;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    EXPECT_LIFE()
    x = ilu_EndUnion(cCall, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szInt64(JIluCall jjcall, Jlong ival)
{
    ilu_longinteger i = TOIluInt64(ival);
    PROLOG_SZ()
    sz = ilu_SizeOfLongInteger(cCall, i, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outInt64(JIluCall jjcall, Jlong ival)
{
    ilu_longinteger i = TOIluInt64(ival);
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputLongInteger(cCall, i, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jlong
xerox_ilu_IluCall_inInt64(JIluCall jjcall)
{
    Jlong result;
    PROLOG(CALL_IN, result)
    ilu_InputLongInteger(cCall, (ilu_longinteger*) &result, &ioerrs);
    ERRCHECK(result)
    return result;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szCard64(JIluCall jjcall, Jlong v)
{
    PROLOG_SZ()
    sz = ilu_SizeOfLongCardinal(cCall, TOIluCard64(v), &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outCard64(JIluCall jjcall, Jlong v)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputLongCardinal(cCall, TOIluCard64(v), &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jlong
xerox_ilu_IluCall_inCard64(JIluCall jjcall)
{
    Jlong result;
    PROLOG(CALL_IN, result)
    ilu_InputLongCardinal(cCall, (ilu_longcardinal*) &result, &ioerrs);
    ERRCHECK(result)
    return result;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szReal64(JIluCall jjcall, Jdouble val)
{
    PROLOG_SZ()
    sz = ilu_SizeOfReal(cCall, (double) val, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outReal64(JIluCall jjcall, Jdouble val)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputReal(cCall, (double) val, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jdouble
xerox_ilu_IluCall_inReal64(JIluCall jjcall)
{
    double value;
    PROLOG(CALL_IN, 0)
    ilu_InputReal(cCall, &value, &ioerrs);
    ERRCHECK(0)
    return (Jdouble) value;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szReal128Buff(JIluCall jjcall, JArrayOfByte jjbytes)
{
    char *buf;
    PROLOG_SZ()
    if (jjbytes==0 OR obj_length(jjbytes)!=sizeof(ilu_longreal)) {
        _ilujava_IluErrorToCallException(0, "xxx_xxx101");
        return 0;
    }
    buf = (char*) unhand(jjbytes)->body;
    /* what about endian-ness? */
    sz = ilu_SizeOfLongReal(cCall, * (ilu_longreal*) buf, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outReal128Buff(JIluCall jjcall, JArrayOfByte jjbytes)
{
    char *buf;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if (jjbytes==0 OR obj_length(jjbytes)!=sizeof(ilu_longreal)) {
        _ilujava_IluErrorToCallException(0, "xxx_xxx103");
        return;
    }
    buf = (char*) unhand(jjbytes)->body;
    /* what about endian-ness? */
    ilu_OutputLongReal(cCall, * (ilu_longreal*) buf, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA void
xerox_ilu_IluCall_inReal128Buff(JIluCall jjcall, JArrayOfByte jjbytes)
{
    char *buf;
    PROLOG(CALL_IN, ILUJAVA_VOID)
    if (jjbytes==0 OR obj_length(jjbytes)!=sizeof(ilu_longreal)) {
        _ilujava_IluErrorToCallException(0, "xxx_xxx104");
        return;
    };
    buf = (char*) unhand(jjbytes)->body;
    /* what about endian-ness? */
    ilu_InputLongReal(cCall, (ilu_longreal*) &buf, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szReal32(JIluCall jjcall, Jfloat f)
{
    PROLOG_SZ()
    sz = ilu_SizeOfShortReal(cCall, (float) f, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outReal32(JIluCall jjcall, Jfloat f)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputShortReal(cCall, (float) f, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jfloat
xerox_ilu_IluCall_inReal32(JIluCall jjcall)
{
    float f;
    PROLOG(CALL_IN, 0)
    ilu_InputShortReal(cCall, &f, &ioerrs);
    ERRCHECK(0)
    return (Jfloat) f;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szEnum(JIluCall jjcall, Jint v)
{
    PROLOG_SZ()
    sz = ilu_SizeOfEnum(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outEnum(JIluCall jjcall, Jint v)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputEnum(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_inEnum(JIluCall jjcall)
{
    ilu_shortcardinal v;
    PROLOG(CALL_IN, 0)
    ilu_InputEnum(cCall, &v, &ioerrs);
    ERRCHECK(0)
    return (Jint) v;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szCard16(JIluCall jjcall, Jshort v)
{
    PROLOG_SZ()
    sz = ilu_SizeOfShortCardinal(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outCard16(JIluCall jjcall, Jshort v)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputShortCardinal(cCall, (ilu_shortcardinal) v, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jshort
xerox_ilu_IluCall_inCard16(JIluCall jjcall)
{
    ilu_shortcardinal v;
    PROLOG(CALL_IN, 0)
    ilu_InputShortCardinal(cCall, &v, &ioerrs);
    ERRCHECK(0)
    return (Jshort) v;
}

EXPORTJAVA Jint
xerox_ilu_IluCall_szCard32(JIluCall jjcall, Jint v)
{
    PROLOG_SZ()
    sz = ilu_SizeOfCardinal(cCall, (ilu_cardinal) v, &ioerrs);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outCard32(JIluCall jjcall, Jint v)
{
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    ilu_OutputCardinal(cCall, (ilu_cardinal) v, &ioerrs);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA Jint
xerox_ilu_IluCall_inCard32(JIluCall jjcall)
{
    ilu_cardinal v;
    PROLOG(CALL_IN, 0)
    ilu_InputCardinal(cCall, &v, &ioerrs);
    ERRCHECK(0)
    return (Jint) v;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szString8(JIluCall jjcall, JString s, Jint limit)
    /* Variable-length array of short character (String) */
{
    Jint len = javaStringLength(s);
    char * cString;
    PROLOG_SZ()
    if ((len > limit) AND (limit != 0)) {
        _ilujava_IluErrorToCallException(0, "string length exceeds limit");
        return 0;
    }
    cString = allocCString(s); 
    	/*(javaString.h) incl. unicode to char conversion */
    sz = ilu_SizeOfString(cCall, cString, 
    	(ilu_cardinal) len, (ilu_cardinal) limit, &ioerrs);
    java_free(cString);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outString8x(JIluCall jjcall, JString s, Jint limit)
    /* Variable-length array of short character (String) */
{
    Jint len = javaStringLength(s);
    char * cString;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if ((len > limit) AND (limit != 0)) {
        _ilujava_IluErrorToCallException(0, "outString8x: length exceeds limit");
        return;
    }
    cString = allocCString(s); 
    	/*(javaString.h) incl. unicode to char conversion */
    if (_ilujava_callsFlag>0) {
       ilu_DebugPrintf("$ outString8x len=%d  s=<%s> \n", len, cString);
    }
    ilu_OutputString(cCall, cString, (ilu_cardinal) len, 
    	(ilu_cardinal) limit, &ioerrs);
    java_free(cString);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA JString
xerox_ilu_IluCall_inString8(JIluCall jjcall, Jint limit)
{
    JString s;
    char * cString = 0;
    ilu_cardinal len = 0;
    PROLOG(CALL_IN, 0)
    ilu_InputString(cCall, &cString, &len, (ilu_cardinal) limit, &ioerrs);
    if (ILU_ERRNOK(ioerrs)) {
        ilu_free(cString);
        _ilujava_IluErrorToException(&ioerrs, "IluCall.inString8");
        return 0;
    }
    if (_ilujava_callsFlag>0) {
        ilu_DebugPrintf("$ inString8 len=%d  s=<%s> \n", len, cString);
    }
    s = makeJavaString(cString, (int) len);
    ilu_free(cString);
    return s;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szBytesSx(JIluCall jjcall, 
    JArrayOfByte jjbytes, 
    Jint start, Jint len, Jint limit
    )
    /* Variable-length array of byte */
{
    char *data;
    Jint dataLen;
    PROLOG_SZ()
    if ((len > limit) AND (limit != 0)) {
        _ilujava_IluErrorToCallException(0, "szBytesSx: length exceeds limit");
        return 0;
    }
    if (jjbytes == 0) {
        data = (char *) &ilujava_nullBuffer;
	dataLen = 0;
	start  = 0;
    } else {
        data = (char *) unhand(jjbytes)->body;
        dataLen = obj_length(jjbytes);
        
    }
    if (start>dataLen OR len>(dataLen-start)) {
        /* data structure too small to contain len bytes */
        _ilujava_IluErrorToCallException(0, "szBytesSx: bad start offset");
        return 0;
    }
    sz = ilu_SizeOfBytes(cCall, (ilu_bytes) data+start, 
    	(ilu_cardinal)len, (ilu_cardinal)limit, 
    	&ioerrs
    	);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outBytesSx(JIluCall jjcall, 
    JArrayOfByte jjbytes, 
    Jint start, Jint len, Jint limit
    )
    /* Variable-length array of byte */
{
    char *data;
    Jint dataLen;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if ((len > limit) AND (limit != 0)) {
        _ilujava_IluErrorToCallException(0, "length exceeds limit");
        return;
    }
    if (jjbytes == 0) {
        data = (char *) &ilujava_nullBuffer;
	dataLen = 0;
	start  = 0;
    } else {
        data = (char *) unhand(jjbytes)->body;
        dataLen = obj_length(jjbytes);
        
    }
    if (start>dataLen OR len>(dataLen-start)) {
        /* data structure too small to contain len jjbytes */
        _ilujava_IluErrorToCallException(0, "outBytesSx: bad start offset");
        return;
    }
    ilu_OutputBytes(cCall, (ilu_bytes) data+start, 
    	(ilu_cardinal)len, (ilu_cardinal)limit, 
    	&ioerrs
    	);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA JArrayOfByte
xerox_ilu_IluCall_inBytesS(JIluCall jjcall, Jint limit)
    /* Variable-length array of byte */
{
    JArrayOfByte jjbytes = 0;
    ilu_bytes data = 0;
    ilu_cardinal len;
    PROLOG(CALL_IN, 0)
    ilu_InputBytes(cCall, &data, 
        &len, (ilu_cardinal) limit, 
    	&ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        _ilujava_IluErrorToException(&ioerrs, "IluCall.inBytesS"); 
        return 0; 
    }
    if (data) {
        jjbytes = IluJava_JAoB_from8((char *)data, len);
        ilu_free(data);
    }
    return jjbytes;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szBytesAx(JIluCall jjcall, 
    JArrayOfByte jjbytes, 
    Jint start, Jint limit
    )
    /* Fixed-length array of byte */
{
    char *data;
    long dataLen;
    long cstart = (long) start;
    long climit = (long) limit;
    PROLOG_SZ()
    if (jjbytes == 0) {
	if (climit>0) {
	    _ilujava_NullPointerException("IluCall");
	    return 0;
	}
	dataLen = 0;
	data = (char *) &ilujava_nullBuffer;
    } else {
        dataLen = obj_length(jjbytes);
        data = (char *) unhand(jjbytes)->body;
    }
    if (cstart>dataLen OR climit>(dataLen-cstart)) {
        _ilujava_IluErrorToCallException(0, "szBytesAx: length exceeds limit");
        return 0;
    }
    sz = ilu_SizeOfOpaque(cCall, (ilu_opaque) data+cstart, 
    	(ilu_cardinal) climit,  
    	&ioerrs
    	);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outBytesAx(JIluCall jjcall, 
    JArrayOfByte jjbytes, 
    Jint start, Jint limit
    )
    /* Fixed-length array of byte */
{
    char *data;
    long dataLen;
    long cstart = (long) start;
    long climit = (long) limit;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if (jjbytes == 0) {
	if (climit>0) {
	    _ilujava_NullPointerException("IluCall");
	    return;
	}
	dataLen = 0;
	data = (char *) &ilujava_nullBuffer;
    } else {
        dataLen = obj_length(jjbytes);
        data = (char *) unhand(jjbytes)->body;
    }
    if (cstart>dataLen OR limit>(dataLen-cstart)) {
        _ilujava_IluErrorToCallException(0, "outBytesAx: length exceeds limit");
        return;
    }
    ilu_OutputOpaque(cCall, (ilu_opaque) data+cstart, 
    	(ilu_cardinal) climit, 
    	&ioerrs
    	);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA JArrayOfByte
xerox_ilu_IluCall_inBytesA(JIluCall jjcall, Jint len)
    /* Fixed-length array of byte */
{
    JArrayOfByte jjbytes;
    ilu_opaque data = 0;
    PROLOG(CALL_IN, 0)
    ilu_InputOpaque(cCall, &data, 
        (ilu_cardinal) len, &ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        _ilujava_IluErrorToException(&ioerrs, "IluCall.inBytesA"); 
        return 0; 
    }
    jjbytes = IluJava_JAoB_from8((char *)data, len);
    ilu_free(data);
    return jjbytes;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szString16(JIluCall jjcall, 
    JString jjstring, 
    Jint limit
    )
    /* Variable-length array of 16 bit character (WString) */
{
    ilu_wstring cw;
    ilu_cardinal climit = limit;
    ilu_cardinal len;
    PROLOG_SZ()
    if (jjstring == 0) {
	len = 0;
    } else {
        len = (ilu_cardinal) javaStringLength(jjstring);
    }
    if ((len>climit) AND (climit!=0)) len = climit;
    cw = ilu_malloc(len*sizeof(ilu_character)+2/*safety*/);
    IluJava_JString_toC16(jjstring, (char *) cw, len);
    sz = ilu_SizeOfWString(cCall, cw, 
    	(ilu_cardinal) len, (ilu_cardinal) climit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK(0)
    return (Jint) sz;
}


EXPORTJAVA void
xerox_ilu_IluCall_outString16x(JIluCall jjcall, 
    JString jjstring, 
    Jint limit
    )
    /* Variable-length array of 16 bit character (WString) */
{
    ilu_wstring cw;
    ilu_cardinal climit = limit;
    ilu_cardinal len;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if (jjstring == 0) jjstring = ilujava_jEmptyString;
    len = (ilu_cardinal) javaStringLength(jjstring);
    if ((len>climit) AND (climit!=0)) len = climit;
    cw = ilu_malloc(len*sizeof(ilu_character)+2/*safety*/);
    IluJava_JString_toC16(jjstring, (char *) cw, len);
    ilu_OutputWString(cCall, cw, 
    	(ilu_cardinal) len, (ilu_cardinal) climit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK(ILUJAVA_VOID)
}


EXPORTJAVA JArrayOfChar
xerox_ilu_IluCall_inString16AsArray(JIluCall jjcall, Jint limit)
    /* Variable-length array of character (WString) */
{
    JArrayOfChar jja;
    ilu_wstring data = 0;
    ilu_cardinal len;
    PROLOG(CALL_IN, 0)
    ilu_InputWString(cCall, &data, 
        &len, (ilu_cardinal) limit, 
    	&ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        _ilujava_IluErrorToException(&ioerrs, "IluCall.inString16"); 
        return 0; 
    }
    jja = IluJava_JAoC_from16((char *)data, len);
    ilu_free(data);
    return jja;
}


EXPORTJAVA Jint
xerox_ilu_IluCall_szChar8Array(JIluCall jjcall, JArrayOfChar jja, Jint lim)
    /* Fix-length array of short character (StringVec) */
{
    Jint len;
    char * cbuffer;
    PROLOG_SZ()
    if (jja == 0) {
	if (lim>0) {
	    _ilujava_NullPointerException("IluCall");
	    return 0;
	}
	len = 0;
	cbuffer = ilu_malloc(2); cbuffer[0] = 0;
    } else {
        len = obj_length(jja);
        if (len<lim) {
            _ilujava_IluErrorToCallException(&ioerrs, "IluCall.szChar8Array");
            return 0;
        }
        cbuffer = ilu_malloc(len + 2 /* 0C + safety */ );
        IluJava_JAoC_toC8(jja, cbuffer, len);
    }
    sz = ilu_SizeOfStringVec(
        cCall, (ilu_string) cbuffer, (ilu_cardinal) lim, &ioerrs);
    ilu_free(cbuffer);
    ERRCHECK(0)
    return (Jint) sz;
} /*szChar8Array*/


EXPORTJAVA void
xerox_ilu_IluCall_outChar8Arrayx(JIluCall jjcall, JArrayOfChar jja, Jint lim)
{
    char * cbuffer;
    Jint len; 
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if (jja == 0) {
	if (lim>0) {
	    _ilujava_NullPointerException("IluCall");
	    return;
	}
	len = 0;
	cbuffer = ilu_malloc(2); cbuffer[0] = 0;
    } else {
        len = obj_length(jja);
        if (len<lim) {
            _ilujava_IluErrorToCallException(&ioerrs, "IluCall.outChar8Arrayx");
            return;
        }
        cbuffer = ilu_malloc(len + 2 /* 0C + safety */ );
        IluJava_JAoC_toC8(jja, cbuffer, len);
    }
    ilu_OutputStringVec(
        cCall, (ilu_string) cbuffer, (ilu_cardinal) lim, &ioerrs);
    ilu_free(cbuffer);
    ERRCHECK(ILUJAVA_VOID)
} /*outChar8Arrayx*/


EXPORTJAVA JArrayOfChar
xerox_ilu_IluCall_inChar8Array(JIluCall jjcall, Jint lim)
{
    JArrayOfChar jja;
    char * cString;
    PROLOG(CALL_IN, 0)
    ilu_InputStringVec(cCall, &cString, (ilu_cardinal) lim, &ioerrs);
    if (ILU_ERRNOK(ioerrs)) {
        ilu_free(cString);
        _ilujava_IluErrorToException(&ioerrs, "IluCall.inChar8Array");
        return 0;
    }
    jja = IluJava_JAoC_from8(cString, lim);
    ilu_free(cString);
    return jja;
} /*inChar8Array*/


EXPORTJAVA Jint
xerox_ilu_IluCall_szChar16Array(JIluCall jjcall, 
    JArrayOfChar jja, 
    Jint limit
    )
    /* Fix-length array of character (WStringVec) */
{
    ilu_wstring cw;
    Jint len;
    PROLOG_SZ()
    if (jja == 0) {
	if (limit>0) {
	    _ilujava_NullPointerException("IluCall");
	    return 0;
	}
	len = 0;
	cw = ilu_malloc(4); cw[0] = 0; cw[1] = 0;
    } else {
        len = obj_length(jja);
        if (len<limit) {
            _ilujava_IluErrorToCallException(&ioerrs, "IluCall.szChar8Array");
            return 0;
        }
        cw = ilu_malloc(limit*sizeof(ilu_character)+4 /*safety+terminator*/ );
        IluJava_JAoC_toC16(jja, (char *)cw, len);
    }
    sz = ilu_SizeOfWStringVec(cCall, cw, 
    	(ilu_cardinal)limit,  
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK(0)
    return (Jint) sz;
} /*szChar16Array*/


EXPORTJAVA void
xerox_ilu_IluCall_outChar16Arrayx(JIluCall jjcall, 
    JArrayOfChar jja, 
    Jint limit
    )
    /* Fix-length array of character (WStringVec) */
{
    ilu_wstring cw;
    Jint len;
    PROLOG(CALL_OUT, ILUJAVA_VOID)
    if (jja == 0) {
	if (limit>0) {
	    _ilujava_NullPointerException("IluCall");
	    return;
	}
	len = 0;
	cw = ilu_malloc(4); cw[0] = 0; cw[1] = 0;
    } else {
        len = obj_length(jja);
        if (len<limit) {
            _ilujava_IluErrorToCallException(&ioerrs, "outChar16Arrayx");
            return;
        }
        cw = ilu_malloc(limit*sizeof(ilu_character)+4 /*safety+terminator*/ );
        IluJava_JAoC_toC16(jja, (char *)cw, len);
    }
    ilu_OutputWStringVec(cCall, cw, 
    	(ilu_cardinal) limit, 
    	&ioerrs
    	);
    ilu_free(cw);
    ERRCHECK(ILUJAVA_VOID)
} /*outChar16Array*/
 

EXPORTJAVA JArrayOfChar
xerox_ilu_IluCall_inChar16Array(JIluCall jjcall, Jint len)
    /* Fix-length array of character (WStringVec) */
{
    JArrayOfChar jja;
    ilu_wstring data = 0;
    PROLOG(CALL_IN, 0)
    ilu_InputWStringVec(cCall, &data, 
        (ilu_cardinal)len, &ioerrs
    	);
    if (ILU_ERRNOK(ioerrs)) {
        if (data) ilu_free(data);
        _ilujava_IluErrorToException(&ioerrs, "IluCall.inChar16Array"); 
        return 0; 
    }
    jja = IluJava_JAoC_from16((char *)data, len);
    ilu_free(data);
    return jja;
} /*inChar16Array*/


EXPORTJAVA void
xerox_ilu_IluCall_nativeMarkCallFailure(JIluCall jjcall, Jint protocolException)
{
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    if (cCall) {
        if (cCall->ca_pe==ilu_ProtocolException_Success)
            cCall->ca_pe = (ilu_ProtocolException) protocolException;
    }
} /*nativeMarkCallFailure*/


/* end */
